/***************************************************************************
 *
 *   ::::_audioeditor_::::
 *   a simple multitrack audio editor
 *
 *   audioctl
 *   source file
 *
 *   begin                : Jan 2006
 *   copyright            : (C) 2006 by Georg Holzmann
 *   email                : grh@mur.at
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 ***************************************************************************/

#include "audioctl.h"
#include "threadutilities.h"
#include <QMessageBox>

// define static members
PortAudioStream *AudioCtl::m_stream = 0;
PaError AudioCtl::m_err = 0;
MutexPos AudioCtl::m_position;
PortAudioCallback *AudioCtl::m_track_callbacks = 0;
QReadWriteLock AudioCtl::m_track_lock;
PortAudioCallback *AudioCtl::m_masterFX_callbacks = 0;
QReadWriteLock AudioCtl::m_masterFX_lock;
bool AudioCtl::m_got_hardware = 0;
int AudioCtl::m_instances = 0;

AudioCtl::AudioCtl()
{
  m_instances++;
  
  // open audio device only at the first creation
  if(m_instances==1)
  {
    m_err = Pa_Initialize();
    if( m_err != paNoError )
    {
      QMessageBox::critical(0, tr("Audio Hardware Error"), 
        tr("Couldn't initialize hardware! <br> Maybe you are running "
           "an other audio application or your hardware settings are incorrect !"));
      m_got_hardware=false;
    }
    else
      m_got_hardware=true;
  
    openPAStream();
  }
}

AudioCtl::~AudioCtl()
{
  // close audio device only at the last destruction
  if(m_instances==1)
  {
    closePAStream();
    Pa_Terminate();
  }
  
  m_instances--;
}

void AudioCtl::play(Position cursor)
{
  if(!m_got_hardware)
    return;
  
  PaError check = Pa_StreamActive(m_stream);
  if(check==1) // allready playing
    return;
  
  if(cursor != m_position.get())
  {
    m_position.set(cursor);
    emit positionChanged( m_position.get() );
  }
  
  m_err = Pa_StartStream(m_stream);
  Q_ASSERT( m_err == paNoError );
}

void AudioCtl::pause()
{
  if(!m_got_hardware)
    return;
  
  PaError check = Pa_StreamActive(m_stream);
  
  if(check==1) // playing
    stop();
  if(check==0) // not playing
    play(m_position.get());
}

void AudioCtl::stop()
{
  m_err = Pa_StopStream(m_stream);
  Q_ASSERT( m_err == paNoError );
}

void AudioCtl::registerTrack(PortAudioCallback *callback)
{
  m_track_lock.lockForWrite();
  m_track_callbacks = callback;
  m_track_lock.unlock();
}

void AudioCtl::registerMasterFX(PortAudioCallback *callback)
{
  m_masterFX_lock.lockForWrite();
  m_masterFX_callbacks = callback;
  m_masterFX_lock.unlock();
}

void AudioCtl::openPAStream(int samplerate, int bufsize)
{
  closePAStream();
  
  if(!m_got_hardware)
    return;
  
  // get parameters for device and open it
  
  //int in_device = Pa_GetDefaultInputDeviceID();
  int out_device = Pa_GetDefaultOutputDeviceID();
  
#if 0
  const PaDeviceInfo *info;
  info = Pa_GetDeviceInfo( out_device );
  QMessageBox::critical(0, tr("out device info"), 
                        tr("struct Version: %1 <br> name: %2 <br> in: %3 <br> out: %4 <br> sr: %5 <br> lo: %6 hi: %7 <br> format: %8 <br> countDevice: %9")
                            .arg(info->structVersion)
                            .arg(info->name)
                            .arg(info->maxInputChannels)
                            .arg(info->maxOutputChannels)
                            .arg(info->numSampleRates)
                            .arg(info->sampleRates[0])
                            .arg(info->sampleRates[1])
                            .arg(info->nativeSampleFormats)
                            .arg(Pa_CountDevices()));
#endif
  
  m_err = Pa_OpenStream(
  &m_stream,                // stream pointer
  paNoDevice,               // input device
  0,                        // nr. of input channels
  paFloat32,                // input sample format
  NULL,                     // input driver info (not needed)
  out_device,               // output device
  2,                        // nr. of output channels
  paFloat32,                // output sample format
  NULL,                     // output driver info (not needed)
  samplerate,               // samplerate
  bufsize,                  // frames per buffer
  1,                        // multibuffered i/o (here only 1 buffer)
  paClipOff | paDitherOff,  // add. stream flags
  mainAudioCallback,        // the audio callback
  NULL );                   // additional custom data
  
  if( m_err != paNoError )
  {
    QMessageBox::critical(0, tr("Audio Hardware Error"), 
                          tr("Invalid parameters for audio device! <br> PortAudio Error: %1!")
                              .arg(Pa_GetErrorText(m_err)));
    m_got_hardware = false;
  }
}

void AudioCtl::closePAStream()
{
  if(!m_got_hardware)
    return;
  
  m_err = Pa_StopStream(m_stream);
  Q_ASSERT( m_err == paNoError );
}

int AudioCtl::mainAudioCallback(void *inputBuffer, void *outputBuffer,
                      unsigned long framesPerBuffer,
                      PaTimestamp outTime, void *userData )
{
  float *out = (float*)outputBuffer;
  
  // STEP 1:
  // first set all samples to zero, so that the tracks can add their data
  for( unsigned int i=0; i<framesPerBuffer; i++ )
  {
    *out++ = 0;  // left
    *out++ = 0;  // right
  }
  
  // STEP 2:
  // call all the track-callbacks, so that they can add their data
  m_track_lock.lockForRead();
  if(m_track_callbacks)
    m_track_callbacks(inputBuffer, outputBuffer, framesPerBuffer,
                      outTime, 0);
  m_track_lock.unlock();
  
  // STEP 3:
  // finally call the Master Effects
  m_masterFX_lock.lockForRead();
  if(m_masterFX_callbacks)
    m_masterFX_callbacks(inputBuffer, outputBuffer, framesPerBuffer,
                         outTime, 0);
  m_masterFX_lock.unlock();
  
  // now make new position
  m_position.add(framesPerBuffer);
  
  return 0;
}
