/***************************************************************************
 *
 *   ::::_audioeditor_::::
 *   a simple multitrack audio editor
 *
 *   audioctl
 *   header file
 *
 *   begin                : Jan 2006
 *   copyright            : (C) 2006 by Georg Holzmann
 *   email                : grh@mur.at
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 ***************************************************************************/

#ifndef AUDIOCTL_H__
#define AUDIOCTL_H__

#include <QObject>
#include <QReadWriteLock>
#include "globaldefines.h"
#include "portaudio.h"

class MutexPos;

/*!
 * \class AudioCtl
 *
 * \brief class for audio IO and synchronisation of tracks and effects
 *
 * AudioCtl uses PortAudio (www.portaudio.com) to access the hardware
 * in a cross-platform way (thanks for it !!!).
 * 
 * The audio IO is callback based: a callback will be called by the 
 * PortAudio engine whenever it needs more audio data for output/input.
 * The callback function operates under an interrupt or background thread.
 * This leaves the foreground application free to do other things while the 
 * audio just runs in the background.
 * 
 * Each track, effect, etc. has to register a REALTIME-SAFE process method
 * in that class and this method will be called each audio block.
 * 
 * AudioCtl manages also all the play/pause/stop/seek and other control
 * methods and synchronizes all different tracks and effects.
 */
class AudioCtl : public QObject
{
  Q_OBJECT
  
  public:
    AudioCtl();
    virtual ~AudioCtl();
    
  public slots:
  
    /*!
     * plays all selected tracks and effects,
     * starting at current position
     * @param cursor the current cursor position
     */
    void play(Position cursor=0);
    
    /*!
     * pauses or unpauses the stream, 
     * like on a CD-player
     */
    void pause();
    
    /*! stops the audio stream */
    void stop();
    
    /*!
     * register the callback of a track
     * it will be called each audio block and can be filled
     * with the data of the track
     * \note the callback function MUST be realtime-safe, so no
     *       memory allocation, fileIO, or other blocking stuff !
     * @param callback a pointer to the callback function
     */
    void registerTrack(PortAudioCallback *callback);
    
    /*!
     * register the callback of an Master Effect
     * it will be called each audio block AFTER all the track callbacks,
     * so you can process global effects here !
     * \note the callback function MUST be realtime-safe, so no
     *       memory allocation, fileIO, or other blocking stuff !
     * @param callback a pointer to the callback function
     */
    void registerMasterFX(PortAudioCallback *callback);
    
  signals:
    
    /*!
     * if the position changes (e.g. seeking in the track)
     * emit a signal, so that you can load buffers for
     * e.g. file IO ...
     * @param new_position the new position in samples
     */
    void positionChanged(Position new_position);
    
  protected:
    
    /*!
     * initializes the audio stream
     * @param samplerate samplerate of the application
     * @param bufsize audio buffer size, if zero then 
     *                use default minimum
     */
    void openPAStream(int samplerate = DEFAULT_SAMPLERATE,
                      int bufsize = DEFAULT_BUFSIZE);
    
    /*! closes the audio stream */
    void closePAStream();
    
  private:
    
    /*!
     * \brief this callback will call all the other track and FX callbacks
     * 
     * This is the main callback function wich gets called from the audio
     * hardware.
     * It will first call all the track callbacks and afterwards the
     * MasterFX callbacks.
     * 
     * @param inputBuffer output sample buffer (size = bufsize * channels)
     * @param outputBuffer input sample buffer (size = bufsize * channels)
     * @param framesPerBuffer number of sample frames to be processed 
     *        by the callback
     * @param outTime time in samples when the buffer(s) processed by
     *        this callback will begin being played at the audio output
     * @param userData value of a user supplied pointer passed to 
     *        Pa_OpenStream(), for storing synthesis data etc.
     * 
     * @return normally 0; nonzero if you want to stop audio
     * 
     * \note as above, this callback must be realtime-save and non of the
     *       other stream functions from PortAUdio may be called from within 
     *       the callback function except for Pa_GetCPULoad()
     *       -> so all locks here are only a HACK for now and 
     *          can/should be removed !
     */
    static int mainAudioCallback(void *inputBuffer, void *outputBuffer,
                      unsigned long framesPerBuffer,
                      PaTimestamp outTime, void *userData );
    
    /*! current position in the tracks in samples, threadsafe! */
    static MutexPos m_position;
    
    /*! calls all the track callbacks, threadsafe! */
    static PortAudioCallback *m_track_callbacks;
    static QReadWriteLock m_track_lock;
    
    /*! calls all the master FX callbacks, threadsafe! */
    static PortAudioCallback *m_masterFX_callbacks;
    static QReadWriteLock m_masterFX_lock;
    
    /*! the portaudio stream */
    static PortAudioStream *m_stream;
    
    /*! for portaudio error messages */
    static PaError m_err;
    
    /*! true if the hardware was initialized sucessfully */
    static bool m_got_hardware;
    
    /*! counts the nr. of instances */
    static int m_instances;
};

#endif // AUDIOCTL_H__
