/***************************************************************************
 *
 *   ::::_audioeditor_::::
 *   a simple multitrack audio editor
 *
 *   threadutilities
 *   header file
 *
 *   begin                : Jan 2006
 *   copyright            : (C) 2006 by Georg Holzmann
 *   email                : grh@mur.at
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 ***************************************************************************/

#ifndef THREADUTILITIES_H__
#define THREADUTILITIES_H__

/*! 
 * \file threadutilities.h
 *
 * some helpers for multithreaded stuff
 * (mainly to synchronize audio and main thread)
 * 
 * \note this all should be lock-free of course :)
 */

#include <QMutex>
#include "globaldefines.h"
#include "portaudio.h"

/*!
 * \class MutexPos
 * 
 * a position with a mutex
 */
class MutexPos
{
  public:
    MutexPos() : m_pos(0) {}
    virtual ~MutexPos() {}
    inline Position get() const { QMutexLocker locker(&m_mutex); return m_pos;  }
    inline void set(Position pos) { QMutexLocker locker(&m_mutex); m_pos = pos; }
    inline void add(Position nr) { QMutexLocker locker(&m_mutex); m_pos += nr; }
    
  protected:
    mutable QMutex m_mutex;
    Position m_pos;
    
  private:
    /*! copy constructor and assignment not allowed */
    MutexPos(const MutexPos &src) { m_pos = src.m_pos; }
    const MutexPos operator = (const MutexPos) { return *this; }
};

#include <QList>
#include <QReadWriteLock>

/*!
 * \class FuncList
 * 
 * a threadsafe function list for PortAudio Callbacks
 * (uses QReadWriteLock: "is a good choice if there are many concurrent reads
 *  and writing occurs infrequently")
 * \todo make it lockfree !
 */
class FuncList
{
  public:
    FuncList() : m_callback(0) {}
    virtual ~FuncList() {}
    
    /*! append element to list */
    inline void append(PortAudioCallback *callback)
    {
      m_lock.lockForWrite();
      m_list.append(callback);
      m_lock.unlock();
    }
    
    /*! removes a specific callback from the list */
    inline void removeAll(PortAudioCallback *callback)
    {
      m_lock.lockForWrite();
      m_list.removeAll(callback);
      m_lock.unlock();
    }
    
    /*! calls all functions in the list */
    inline void callAll(void *inputBuffer, void *outputBuffer,
                        unsigned long framesPerBuffer,
                        PaTimestamp outTime, void *userData )
    {
      m_lock.lockForRead();
      foreach (m_callback, m_list)
        m_callback(inputBuffer, outputBuffer, framesPerBuffer, outTime, userData);
      m_lock.unlock();
    }
    
  protected:
    QReadWriteLock m_lock;
    QList<PortAudioCallback *> m_list;
    PortAudioCallback *m_callback;
    
  private:
    /*! copy constructor and assignment not allowed */
    FuncList(const FuncList &src) { m_list = src.m_list; }
    const FuncList operator = (const FuncList) { return *this; }
};

#endif // THREADUTILITIES_H__
